/*
 * Copyright (c) 2012-2014,2016-2019,2022-2023 LAAS/CNRS
 * All rights reserved.
 *
 * Redistribution and use  in source  and binary  forms,  with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice and this list of conditions.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice and  this list of  conditions in the  documentation and/or
 *      other materials provided with the distribution.
 *
 *					Anthony Mallet on Wed Mar 14 2012
 */
#include <sys/types.h>
#include <assert.h>
#include <errno.h>
#include <err.h>
#include <fcntl.h>
#ifdef HAVE_FEATURES_H
# include <features.h>
#endif
#include <getopt.h>
#include <pthread.h>
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "csLib.h"
#include "h2devLib.h"
#include "h2initGlob.h"
#include "semLib.h"
#include "taskLib.h"

#include "genom3/c/client.h"
#include "genom-iface.h"
#include "private.h"


/* --- local data ---------------------------------------------------------- */

static int pocolibs_h2init = 0;
static int pocolibs_tdev = ERROR;
static size_t pocolibs_mboxsize = 0;
static int ntasks;

static pthread_t pocolibs_evtid;
static int pocolibs_evtid_init = 0;
static pthread_mutex_t evmutex = PTHREAD_MUTEX_INITIALIZER;

static genom_client clients = NULL;

static genom_event	pocolibs_newrqstlk(genom_client h, size_t outsz,
                                int *rqstid);
static genom_event	pocolibs_waitlk(genom_client h, int rqstid);
static genom_event	pocolibs_cleanlk(genom_client h, int rqstid);
static int		pocolibs_recvglk(genom_client h);
static genom_event	pocolibs_doeventslk(genom_client h);

static void *		pocolibs_evtask(void *data);
static void *		genom_pocolibs_context_alloc(genom_context self);
static void		genom_pocolibs_context_free(void *data);


/* --- pocolibs_newclient -------------------------------------------------- */

genom_client
pocolibs_newclient()
{
  genom_client h = malloc(sizeof(*h));
  if (!h) return NULL;

  h->context.raise = genom_pocolibs_client_raise;
  h->context.raised = genom_pocolibs_client_raised;
  h->context.data = &h->context_data;
  if (pthread_key_create(&h->context_data.key, genom_pocolibs_context_free))
    goto error;

  h->instance = NULL;
  h->csid = NULL;
  pthread_mutex_init(&h->cmutex, NULL);
  pthread_cond_init(&h->cevent, NULL);
  h->active_rqst = NULL;
  h->nactive = h->nextrqst = 0;
  h->ports = NULL;

  h->evpending = 0;
  h->evpipe[0] = h->evpipe[1] = -1;

  pthread_mutex_lock(&evmutex);
  h->next = clients;
  clients = h;
  pthread_mutex_unlock(&evmutex);

  return h;
error:
  errnoSet(errno);
  free(h);
  return NULL;
}


/* --- pocolibs_delclient -------------------------------------------------- */

void
pocolibs_delclient(genom_client h)
{
  if (!h) return;

  pthread_mutex_lock(&evmutex);
  if (h == clients) {
    clients = h->next;
  } else {
    genom_client c;

    for(c = clients; c && c->next != h; c = c->next);
    if (c) c->next = h->next;
  }
  pthread_mutex_unlock(&evmutex);

  /* context check for mboxLib association */
  if (!pthread_getspecific(h->context_data.key))
    genom_pocolibs_context_alloc(&h->context);

  if (h->csid) csClientEnd(h->csid);
  pocolibs_fini();
  pthread_key_delete(h->context_data.key);
  pthread_mutex_destroy(&h->cmutex);
  pthread_cond_destroy(&h->cevent);
  if (h->instance) free(h->instance);
  if (h->active_rqst) free(h->active_rqst);
  if (h->ports) free(h->ports);

  if (h->evpipe[0] >= 0) close(h->evpipe[0]);
  if (h->evpipe[1] >= 0) close(h->evpipe[1]);

  free(h);
}


/* --- pocolibs_init ------------------------------------------------------- */

genom_event
pocolibs_init(genom_client h, size_t mboxsize)
{
  char name[H2_DEV_MAX_NAME];
  STATUS s;
  int e;

  /* this is not reentrant */
  pthread_mutex_lock(&evmutex);

  /* general pocolibs init */
  if (!pocolibs_h2init) {
    /* application must not have done that */
    if (h2initGlob(0) != OK) goto error;
    pocolibs_h2init = 1;
  }

  /* mailbox init */
  if (pocolibs_mboxsize == 0) {
    /* the first caller creates a reception mailbox */
    snprintf(name, sizeof(name), "cclient-%d", getpid());
    s = csMboxInit(name, 0, mboxsize);
    if (s != OK) {
      genom_mwerr_detail d;
      char *p;
      e = errnoGet();
      p = d.what;
      p += snprintf(p, sizeof(d.what), "%s: ", name);
      h2getErrMsg(e, p, sizeof(d.what) - (p - d.what));
      pthread_mutex_unlock(&evmutex);
      return genom_mwerr(&d, &h->context);
    }
    pocolibs_mboxsize = mboxsize;
    pocolibs_tdev = taskGetUserData(0);
  }

  /* context init */
  if (!pthread_getspecific(h->context_data.key))
    genom_pocolibs_context_alloc(&h->context);

  /* update mailbox size if needed */
  if (pocolibs_mboxsize < mboxsize) {
    s = csMboxUpdate(0, mboxsize);
    if (s != OK) goto error;
    pocolibs_mboxsize = mboxsize;
  }

  /* create event task */
  if (!pocolibs_evtid_init) {
    s = pthread_create(&pocolibs_evtid, NULL, pocolibs_evtask, NULL);
    if (s) { errnoSet(errno); goto error; }
  }
  pocolibs_evtid_init++;

  pthread_mutex_unlock(&evmutex);
  return genom_ok;

error:
  pthread_mutex_unlock(&evmutex);
  e = errnoGet();
  if (H2_SYS_ERR_FLAG(e))
    return genom_syserr(&(genom_syserr_detail){ .code = e }, &h->context);
  else {
    genom_mwerr_detail d;
    h2getErrMsg(e, d.what, sizeof(d.what));
    return genom_mwerr(&d, &h->context);
  }
  assert(!"not reached");
}


/* --- pocolibs_fini ------------------------------------------------------- */

void
pocolibs_fini(void)
{
  pthread_mutex_lock(&evmutex);
  if (pocolibs_evtid_init) pocolibs_evtid_init--;
  if (pocolibs_evtid_init) {
    pthread_mutex_unlock(&evmutex);
    return;
  }

  if (pocolibs_tdev != ERROR)
    h2semGive(H2DEV_TASK_SEM_ID(pocolibs_tdev));

  if (pocolibs_mboxsize)
    csMboxEnd();
  pocolibs_mboxsize = 0;
  pocolibs_tdev = ERROR;
  pthread_mutex_unlock(&evmutex);

  pthread_join(pocolibs_evtid, NULL);
}


/* --- pocolibs_getopt ----------------------------------------------------- */

/* options */
static const char shortopts_string[] = "+i:";
static struct option longopts_list[] = {
  { "name",		required_argument,	NULL,	'i'},
  { "maxmsg",		required_argument,	NULL,	-'m'},
  { NULL, 0, NULL, 0}
};

const struct genom_client_option genom_pocolibs_options[] = {
  { "-i|--name",	"instance name" },
  { "--maxmsg",		"maximum message size" },
  { NULL, NULL}
};

genom_event
pocolibs_getopt(genom_client h, int argc, char *argv[],
                struct pocolibs_options *opts)
{
  int c;
  char *e;

  opts->varmsg_maxsize = 32768;
  opts->instance = NULL;

  /* reset getopt */
#ifdef HAVE_OPTRESET
  optreset = 1;
#endif
#ifdef __GLIBC__
  optind = 0; /* not "1", see gnu manual for getopt(3) ... */
#else
  optind = 1;
#endif

  /* parse command line options */
  opterr = 0;
  while (
    (c = getopt_long(argc, argv, shortopts_string, longopts_list, NULL)) != -1)
    switch (c) {
      case 0: break;

      case 'i': opts->instance = optarg; break;
      case -'m': {
        unsigned long s = strtoul(optarg, &e, 0);
        if (*e != 0 || s <= 0)
          return genom_syserr(&(genom_syserr_detail){.code = EINVAL},
                              &h->context);
        opts->varmsg_maxsize = s * 1024;
        break;
      }

      case '?':
      default:
        c = '?'; break;
    }
  if (argc > optind) c = '?';
  if (c == '?')
    return genom_syserr(&(genom_syserr_detail){.code = EINVAL}, &h->context);
  return genom_ok;
}


/* --- pocolibs_newrqst ---------------------------------------------------- */

static genom_event
pocolibs_newrqstlk(genom_client h, size_t outsz, int *rqstid)
{
  struct genom_client_request *a;
  int i;
  assert(h);

  *rqstid = -1;
  for(i = h->nextrqst; i < h->nextrqst + h->nactive; i++)
    if (h->active_rqst[i%h->nactive].status == FREE) {
      *rqstid = i%h->nactive;
      break;
    }
  if (*rqstid < 0) {
    a = realloc(h->active_rqst, (h->nactive+1)*sizeof(*a));
    if (!a)
      return genom_syserr(&(genom_syserr_detail){.code = errno}, &h->context);
    *rqstid = h->nactive;
    h->nactive++;
    h->active_rqst = a;
  }
  h->nextrqst = *rqstid + 1;
  h->evpending = 1;

  h->active_rqst[*rqstid] = (struct genom_client_request){
    .rid = -1,
    .aid = -1,
    .status = FREE,
    .csstatus = FREE,
    .report = genom_ok,
    .output = malloc(outsz)
  };

  if (!h->active_rqst[*rqstid].output)
    return genom_syserr(&(genom_syserr_detail){.code = errno}, &h->context);
  return genom_ok;
}


/* --- pocolibs_eventfd ---------------------------------------------------- */

int
pocolibs_eventfd(genom_client h)
{
  int fd;

  pthread_mutex_lock(&h->cmutex);
  if (h->evpipe[0] < 0) {
    if (pipe(h->evpipe) || fcntl(h->evpipe[0], F_SETFL, O_NONBLOCK))
      genom_syserr(&(genom_syserr_detail){.code = errno}, &h->context);
  }

  fd = h->evpipe[0];
  pthread_mutex_unlock(&h->cmutex);
  return fd;
}


/* --- pocolibs_sendrqst --------------------------------------------------- */

genom_event
pocolibs_sendrqst(genom_client h, int rqst,
                  const struct genom_service_info *info,
                  const void *in, int (*encode)(char *, int, char *, int),
                  int ack, void (*outini)(void *),
                  int (*decode)(char *, int, char *, int), size_t outsz,
                  genom_request_cb sentcb, genom_request_cb donecb,
                  void *cb_data, int *rqstid)
{
  genom_event ex;
  int rid, e;

  /* context check for mboxLib association */
  if (!pthread_getspecific(h->context_data.key))
    genom_pocolibs_context_alloc(&h->context);

  /* gcomLib needs exclusive access for sending (array of sent letters).
   * Acquire evmutex here before client mutex to prevent deadlock. */
  pthread_mutex_lock(&evmutex);

  /* new client request */
  pthread_mutex_lock(&h->cmutex);
  ex = pocolibs_newrqstlk(h, outsz, rqstid);
  if (ex) {
    pthread_mutex_unlock(&h->cmutex);
    pthread_mutex_unlock(&evmutex);
    return ex;
  }

  /* send */
  e = csClientRqstSend(h->csid, rqst, (char *)in, 0, encode, ack, 0, 0, &rid);
  if (e != OK && errnoGet() == S_csLib_TOO_MANY_RQST_IDS) {
    /* The doevents function may not have been called recently enough by a
     * client, or the event thread may not have been scheduled in time. In any
     * case we have no room for more requests. It is worth trying to process
     * any pending events that may empty the mailbox.
     * Safe to do it on behalf of the evtask as the evmutex is still held */
    pocolibs_recvglk(h);

    e = csClientRqstSend(h->csid, rqst, (char *)in, 0, encode, ack, 0, 0, &rid);
  }

  /* done with gcomLib */
  pthread_mutex_unlock(&evmutex);

  /* fill request information */
  if (e != OK) {
    e = errnoGet();

    free(h->active_rqst[*rqstid].output);
    pthread_mutex_unlock(&h->cmutex);

    switch(e) {
      case S_csLib_TOO_MANY_RQST_IDS:
        return genom_too_many_activities(&h->context);

      default: {
        genom_mwerr_detail d;
        char msg[sizeof(d.what)];
        h2getErrMsg(e, msg, sizeof(msg));
        if (snprintf(d.what, sizeof(d.what),
                     "%s csClientRqstSend: %s", h->instance, msg) < 0)
          strncpy(d.what, "internal error", sizeof(d.what));
        return genom_mwerr(&d, &h->context);
      }
    }
    /*NOTREACHED*/
  }

  outini(h->active_rqst[*rqstid].output);
  h->active_rqst[*rqstid].info = info;
  h->active_rqst[*rqstid].rid = rid;
  h->active_rqst[*rqstid].decode = decode;
  h->active_rqst[*rqstid].sentcb = sentcb;
  h->active_rqst[*rqstid].donecb = donecb;
  h->active_rqst[*rqstid].cb_data = cb_data;
  h->active_rqst[*rqstid].status = h->active_rqst[*rqstid].csstatus =
    ack ? WAITING_INTERMED_REPLY : WAITING_FINAL_REPLY;

  pthread_mutex_unlock(&h->cmutex);
  return genom_ok;
}


/* --- pocolibs_actinfo ---------------------------------------------------- */

const struct genom_service_info *
pocolibs_service_info(genom_client h, int rqstid)
{
  const struct genom_service_info *info;

  pthread_mutex_lock(&h->cmutex);
  if (rqstid < 0 || rqstid >= h->nactive) info = NULL;
  else if (h->active_rqst[rqstid].status == FREE) info = NULL;
  else info = h->active_rqst[rqstid].info;
  pthread_mutex_unlock(&h->cmutex);

  return info;
}


/* --- pocolibs_wait ------------------------------------------------------- */

genom_event
pocolibs_wait(genom_client h, int rqstid)
{
  genom_event e;

  pthread_mutex_lock(&h->cmutex);
  e = pocolibs_waitlk(h, rqstid);
  pthread_mutex_unlock(&h->cmutex);
  return e;
}

static genom_event
pocolibs_waitlk(genom_client h, int rqstid)
{
  genom_event e;

  if (rqstid < 0 || rqstid >= h->nactive)
    return genom_syserr(&(genom_syserr_detail){.code = ENOENT}, &h->context);

  while(h->active_rqst[rqstid].csstatus == WAITING_INTERMED_REPLY ||
        h->active_rqst[rqstid].csstatus == WAITING_FINAL_REPLY) {
    pthread_cond_wait(&h->cevent, &h->cmutex);

    e = pocolibs_doeventslk(h);
    if (e) return e;
  }

  return genom_ok;
}


/* --- pocolibs_done ------------------------------------------------------- */

int
pocolibs_done(genom_client h, int rqstid)
{
  if (rqstid < 0 || rqstid >= h->nactive) return 1;
  if (h->active_rqst[rqstid].csstatus == WAITING_INTERMED_REPLY ||
      h->active_rqst[rqstid].csstatus == WAITING_FINAL_REPLY)
    return 0;
  return 1;
}


/* --- pocolibs_result ----------------------------------------------------- */

genom_event
pocolibs_result(genom_client h, int rqstid,
                genom_event *report, void **output, void **exdetail)
{
  struct genom_client_request *a;
  genom_event s;

  pthread_mutex_lock(&h->cmutex);
  s = pocolibs_waitlk(h, rqstid);
  if (s) {
    pthread_mutex_unlock(&h->cmutex);
    return s;
  }

  a = &h->active_rqst[rqstid];
  if (a->status == FREE) {
    pthread_mutex_unlock(&h->cmutex);
    return genom_syserr(&(genom_syserr_detail){.code = ENOENT}, &h->context);
  }

  if (report) *report = a->report;
  if (output) *output = a->output;
  if (exdetail) *exdetail = a->exdetail;
  pthread_mutex_unlock(&h->cmutex);
  return genom_ok;
}


/* --- pocolibs_clean ------------------------------------------------------ */

genom_event
pocolibs_clean(genom_client h, int rqstid)
{
  genom_event e;

  pthread_mutex_lock(&h->cmutex);
  e = pocolibs_cleanlk(h, rqstid);
  pthread_mutex_unlock(&h->cmutex);

  return e;
}

static genom_event
pocolibs_cleanlk(genom_client h, int rqstid)
{
  struct genom_client_request *a;

  if (rqstid < 0 || rqstid >= h->nactive)
    return genom_syserr(&(genom_syserr_detail){.code = ENOENT}, &h->context);

  a = &h->active_rqst[rqstid];
  if (a->status == FREE)
    return genom_syserr(&(genom_syserr_detail){.code = ENOENT}, &h->context);

  if (a->csstatus != FINAL_REPLY_OK) {
    /* don't clear the request yet: it might be in a callback invocation or
     * still waiting replies */
    a->status = FINAL_REPLY_TIMEOUT;
    return genom_ok;
  }

  a->status = FREE;
  if (a->output) {
    if (a->info->fini_output) a->info->fini_output(a->output);
    free(a->output);
    a->output = NULL;
  }
  if (a->exdetail) {
    assert(a->report);
    if (a->exfini) a->exfini(a->exdetail);
    free(a->exdetail);
    a->exdetail = NULL;
  }

  return genom_ok;
}


/* --- pocolibs_recv ------------------------------------------------------- */

static int
pocolibs_recvglk(genom_client h)
{
  struct genom_client_request *a;
  int i, s;
  int new;

  /* check all pending requests */
  for(i = new = 0; i < h->nactive; i++) {
    a = &h->active_rqst[i];
    if (a->status == FREE ||
        a->csstatus == FREE ||
        a->csstatus == FINAL_REPLY_OK) continue;

    assert(a->exdetail == NULL);

    s = csClientReplyRcv(
      h->csid, a->rid, NO_BLOCK,
      (char *)&a->aid, sizeof(a->aid), NULL, (void *)a, 0, a->decode);
    if (a->csstatus == s) continue;

    /* new reply */
    new = 1;
    switch(s) {
      case WAITING_INTERMED_REPLY:
      case WAITING_FINAL_REPLY:
      case FINAL_REPLY_OK:
        /* valid reply */
        a->csstatus = s;
        break;

      default: /* == ERROR */
        a->csstatus = FINAL_REPLY_OK;
        if (a->report == genom_ok) {
          /* valid request reply means middleware error */
          a->report = genom_mwerr_id;
          a->exfini = NULL;
          a->exdetail = malloc(sizeof(genom_mwerr_detail));
          if (a->exdetail) {
            genom_mwerr_detail *d = a->exdetail;
            h2getErrMsg(errnoGet(), d->what, sizeof(d->what));
            d->what[sizeof(d->what)-1] = *"";
          } else
            a->report = genom_fatal;
        }
    }
  }

  return new;
}


/* --- pocolibs_doevents --------------------------------------------------- */

genom_event
pocolibs_doevents(genom_client h)
{
  genom_event e;

  pthread_mutex_lock(&h->cmutex);
  e = pocolibs_doeventslk(h);
  pthread_mutex_unlock(&h->cmutex);
  return e;
}

static genom_event
pocolibs_doeventslk(genom_client h)
{
  struct genom_client_request *a;
  int evpending;
  char c[16];
  ssize_t r;
  int i;

  /* flush pipe */
  if (h->evpipe[0] >= 0)
    do {
      do {
        r = read(h->evpipe[0], c, sizeof(c));
      } while (r == -1 && errno == EINTR);
      switch(r) {
        case 0:
          return genom_syserr(
            &(genom_syserr_detail){.code = EBADF}, &h->context);

        case -1:
          if (errno != EAGAIN && errno != EWOULDBLOCK)
            return genom_syserr(
              &(genom_syserr_detail){.code = errno}, &h->context);
          break;
      }
    } while(r == sizeof(c));

  /* update statuses */
  evpending = 0;
  for(i = 0; i < h->nactive; i++) {
    a = &h->active_rqst[i];
    if (a->status == FREE || a->csstatus == FREE) continue;
    if (a->status != FINAL_REPLY_OK) evpending = 1;
    if (a->status == FINAL_REPLY_TIMEOUT) {
      /* the request was cleaned before the final reply: discard it */
      if (a->csstatus == FINAL_REPLY_OK) pocolibs_cleanlk(h, i);
      continue;
    }
    if (pocolibs_tdev == ERROR && a->csstatus != FINAL_REPLY_OK) {
      /* evtask left on error */
      a->csstatus = FINAL_REPLY_OK;
      a->report = genom_fatal;
    }

    if (a->status == a->csstatus) continue;
    a->status = a->csstatus;

    /* invoke callbacks - this must be done with lock released, as the
     * callback is allowed to send new requests or clean existing ones */
    switch(a->status) {
      case WAITING_FINAL_REPLY:
        if (a->sentcb) {
          pthread_mutex_unlock(&h->cmutex);
          (*a->sentcb)(h, i, 0, genom_ok, NULL, NULL, a->cb_data);
          pthread_mutex_lock(&h->cmutex);
        }
        break;

      case FINAL_REPLY_OK:
        if (a->donecb) {
          a->csstatus = FREE; /* prevent cleaning */
          pthread_mutex_unlock(&h->cmutex);
          (*a->donecb)(h, i, 1, a->report, a->output, a->exdetail, a->cb_data);
          pthread_mutex_lock(&h->cmutex);
          a->csstatus = FINAL_REPLY_OK;
          /* the request was cleaned: discard it */
          if (a->status == FINAL_REPLY_TIMEOUT) pocolibs_cleanlk(h, i);
        }
        break;

      default: assert(!"reached unreachable");
    }
  }
  if (!evpending) h->evpending = 0;

  return genom_ok;
}


/* --- pocolibs_evtask ----------------------------------------------------- */

static void *
pocolibs_evtask(void *data)
{
  genom_client h;
  H2SEM_ID sem;
  ssize_t s;

  /* mboxLib mailbox association */
  pthread_mutex_lock(&evmutex);
  if (pocolibs_tdev == ERROR) goto done;
  genom_pocolibs_context_alloc(NULL);
  sem = H2DEV_TASK_SEM_ID(pocolibs_tdev);
  pthread_mutex_unlock(&evmutex);

  while (h2semTake(sem, 0) == TRUE) {
    if (!pocolibs_evtid_init) break;

    pthread_mutex_lock(&evmutex);
    for(h = clients; h; h = h->next) {
      pthread_mutex_lock(&h->cmutex);

      /* receive all pending requests replies */
      if (h->evpending && pocolibs_recvglk(h)) {
        /* signal any waiting thread */
        pthread_cond_broadcast(&h->cevent);

        /* signal fd-based events */
        if (h->evpipe[1] >= 0)
          do
            s = write(h->evpipe[1], &(char){0}, 1);
          while (s < 0 && errno == EINTR);
      }

      pthread_mutex_unlock(&h->cmutex);
    }
    pthread_mutex_unlock(&evmutex);
  }

  pthread_mutex_lock(&evmutex);
  pocolibs_tdev = ERROR;
done:
  for(h = clients; h; h = h->next) {
    pthread_mutex_lock(&h->cmutex);
    /* signal any waiting thread */
    pthread_cond_broadcast(&h->cevent);

    /* signal fd-based events */
    if (h->evpipe[1] >= 0)
      do
        s = write(h->evpipe[1], &(char){0}, 1);
      while (s < 0 && errno == EINTR);
    pthread_mutex_unlock(&h->cmutex);
  }
  pthread_mutex_unlock(&evmutex);
  return NULL;
}


/* --- exceptions ---------------------------------------------------------- */

genom_event
genom_pocolibs_client_raise(genom_event ex, void *detail, size_t size,
                            genom_context self)
{
  struct genom_exdata *data = pthread_getspecific(self->data->key);

  if (!data) {
    data = genom_pocolibs_context_alloc(self);
    if (!data) return genom_fatal;
  }

  if (data->exsize < size) {
    void *exdetail = realloc(data->exdetail, size);
    if (!exdetail) return data->ex = genom_fatal;

    data->exsize = size;
    data->exdetail = exdetail;
  }
  data->ex = ex;
  memcpy(data->exdetail, detail, size);
  return ex;
}

const void *
genom_pocolibs_client_raised(genom_event *ex, genom_context self)
{
  struct genom_exdata *data = pthread_getspecific(self->data->key);

  if (ex) *ex = data ? data->ex : genom_ok;
  return data ? data->exdetail : NULL;
}

static void *
genom_pocolibs_context_alloc(genom_context self)
{
  struct genom_exdata *c;
  long tid;

  /* ensure that caller thread is a pocolibs task - although this is not
   * related to genom context, this is done here as a convenience, as this
   * function is typically invoked once per thread */
  tid = taskIdSelf();
  if (!tid) {
    char name[H2_DEV_MAX_NAME];
    snprintf(name, sizeof(name), "cclient-%d/%d", (int)getpid(), ++ntasks);
    tid = taskFromThread(name);
    if (tid == ERROR) return NULL;

    /* mboxLib mailbox association - this is cheating with pocolibs, but it
     * should be OK as long as the current pocolibs design does not change */
    taskSetUserData(tid, pocolibs_tdev);
  }

  /* when invoked during event task initialization, there is no context */
  if (!self) return NULL;

  /* create exception context */
  c = malloc(sizeof(*c));
  if (!c) return NULL;

  c->ex = genom_ok;
  c->exdetail = NULL;
  c->exsize = 0;

  /* register data */
  if (pthread_setspecific(self->data->key, c)) {
    free(c);
    return NULL;
  }

  return c;
}

static void
genom_pocolibs_context_free(void *data)
{
  struct genom_exdata *c = data;

  if (c->exdetail) free(c->exdetail);
  free(data);
}
